from __future__ import annotations

import os
import sys
import typing as t

import arc
from arc.define.param import groups
import arc.typing as at
from arc import errors
from arc.logging import WARNING, logger, mode_map, DEBUG
from arc.runtime.init import InitMiddleware
from arc.runtime.middleware import Middleware, MiddlewareManager
from arc.runtime.plugin import PluginManager

if t.TYPE_CHECKING:
    from arc.define import Command


class App(MiddlewareManager):
    def __init__(
        self,
        root: Command,
        init_middlewares: t.Sequence[Middleware] = None,
        state: dict[str, t.Any] = None,
        ctx: dict[str, t.Any] = None,
    ) -> None:
        super().__init__(init_middlewares or InitMiddleware.all())
        self.root = root
        self.provided_ctx = ctx or {}
        self.state = state or {}
        self.config = root.config
        self.plugins = PluginManager()

    def __call__(self, input: at.InputArgs = None) -> t.Any:
        self._handle_dynamic_name()
        self._setup_logger()
        ctx = self._create_ctx({"arc.input": input})
        try:
            try:
                ctx = self._stack.start(ctx)
                if "arc.command" not in ctx:
                    raise errors.CommandError(
                        "The command was not decided upon during initialization "
                        "(ctx['arc.command'] is not set). This likely means there "
                        "is a problem with the middleware stack"
                    )

                command: Command = ctx["arc.command"]
                res = None
                res = command.run(ctx)
            except Exception as e:
                res = None
                self._stack.throw(e)
            else:
                res = self._stack.close(res)
        except errors.ArcError as exc:
            if self.config.environment == "production":
                arc.info(exc.fmt(ctx))
                arc.exit(1)

            raise
        except errors.Exit as exc:
            if exc.message:
                arc.info(exc.fmt(ctx))
            raise

        return res

    def execute(self, command: Command, **kwargs: t.Any) -> t.Any:
        args = self._unroll_param_groups(kwargs)
        ctx = self._create_ctx({"arc.command": command, "arc.parse.result": args})
        return command.run(ctx)

    def _create_ctx(self, data: dict[str, t.Any] = None) -> arc.Context:
        return arc.Context(
            {
                "arc.root": self.root,
                "arc.config": self.config,
                "arc.errors": [],
                "arc.app": self,
                "arc.state": self.state,
                "arc.logger": logger,
                "arc.plugins": self.plugins,
            }
            | self.provided_ctx
            | (data or {})
        )

    def _handle_dynamic_name(self) -> None:
        if not self.root.explicit_name:
            name = sys.argv[0]
            self.root.name = os.path.basename(name)

    def _setup_logger(self) -> None:
        if self.config.debug:
            logger.setLevel(DEBUG)
        else:
            logger.setLevel(mode_map.get(self.config.environment, WARNING))

    def _unroll_param_groups(self, args: dict[str, t.Any]) -> dict[str, t.Any]:
        flattened_args: dict[str, t.Any] = {}

        for key, value in args.items():
            if groups.isgroup(value):
                flattened_args |= self._unroll_param_groups(vars(value))
            else:
                flattened_args[key] = value

        return flattened_args
